/**
*  drivers/mtd/nand/nand512w3a.c
*
*  Copyright (C) 2007 MSc
*
*  Derived from drivers/mtd/nand/h1910.c
*       Copyright (C) 2002 Marius Grger (mag@sysgo.de)
*       Copyright (c) 2001 Thomas Gleixner (gleixner@autronix.de)
*
* $Id: h1910.c,v 1.6 2005/11/07 11:14:30 gleixner Exp $
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
*  Overview:
*   This is a device driver for the NAND flash device found on the
*   i12 board which utilizes the ST nand512w3a part. This is
*   a 512Mibit (64MiB x 8 bits) NAND flash device.
*
*  Pin assignment on the pio:
*  0...CLE...Command latch enable, active high
*  1...ALE...Address latch enable, active high
*  2...NCE...Chip enable, active low
*  3...NWP...Write protect, active low
*  4...NU....Not used, always 0
*  5...NU....Not used, always 0
*  6...NU....Not used, always 0
*  7...RB....Ready/Busy, Ready=1,Busy=0
*
*  3/2008: Modified for Kernel 2.6.23
*  Writing to chip is now done directly in nand512w3a_hwcontrol
*
*/

#include <linux/slab.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/nios.h>

/*
 * MTD structure for i12 board
 */
static struct mtd_info *nand512w3a_nand_mtd = NULL;

/*
 * Module stuff
 */

#ifdef CONFIG_MTD_PARTITIONS

/*
 * Define static partitions for flash device
 */
#define NUM_PARTITIONS 2
static struct mtd_partition partition_info[] = {
	{ name: "nand512w3a NAND Flash Config",
		  offset: 0,
		  size: 1*1024*1024 
	},
	{ name: "nand512w3a NAND Flash Image",
		  offset: 0x100000,
		  size: 63*1024*1024 
	}
};

#endif

/**
 * nand_write_byte - [DEFAULT] write one byte to the chip
 * @mtd:	MTD device structure
 * @byte:	pointer to data byte to write
 *
 * Default write function for 8it buswith
 */
static void nand_write_byte(struct mtd_info *mtd, u_char byte)
{
	struct nand_chip *this = mtd->priv;
	writeb(byte, this->IO_ADDR_W);
}


/**
* Write 1 Byte to the pio
*/
void pio_nand_flash_write(unsigned char cValue)
{
  np_pio *pio=(np_pio*)(na_nand_flash_pio);

  outl(-1,&pio->np_piodirection);
  outl(cValue,&pio->np_piodata);
}

/**
* Read 1 Byte from the pio
*/
unsigned char pio_nand_flash_read(void)
{
  np_pio *pio=(np_pio*)(na_nand_flash_pio);
  unsigned char cHelp;

  outl(0,&pio->np_piodirection);
  cHelp=(unsigned char)inl(&pio->np_piodata);
  return(cHelp);
}

/**
* CLE...command latch enable, high active
*/
void pio_nand_flash_set_cle(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp | 0x09;                        //also set wp pin
  pio_nand_flash_write(cHelp);
}

void pio_nand_flash_clr_cle(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp & 0xFE;
  pio_nand_flash_write(cHelp);
}

/**
* ALE...address latch enable, high active
*/
void pio_nand_flash_set_ale(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp | 0x0A;                        //also set wp pin
  pio_nand_flash_write(cHelp);
}

void pio_nand_flash_clr_ale(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp & 0xFD;
  pio_nand_flash_write(cHelp);
}

/**
* CE...chip enable, low active
*/
void pio_nand_flash_set_nce(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp & 0xFB;
  pio_nand_flash_write(cHelp);
}

void pio_nand_flash_clr_nce(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp | 0x0C;                        //also set wp pin
  pio_nand_flash_write(cHelp);
}

/**
* WP...write protect, low active
*/
void pio_nand_flash_set_nwp(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp & 0xF7;
  pio_nand_flash_write(cHelp);
}

void pio_nand_flash_clr_nwp(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp | 0x08;
  pio_nand_flash_write(cHelp);
}

/*
 *	hardware specific access to control-lines
 */
static void nand512w3a_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
  if(ctrl & NAND_CTRL_CHANGE)
  {
 
    if(ctrl & NAND_NCE)
    {
      pio_nand_flash_set_nce();
    }
    else
    {
      pio_nand_flash_clr_nce();
    }
    
    if(ctrl & NAND_CLE)
    {
      pio_nand_flash_set_cle();
    }
    else
    {
      pio_nand_flash_clr_cle();
    }
    
    if(ctrl & NAND_ALE)
    {
      pio_nand_flash_set_ale();
    }
    else
    {
      pio_nand_flash_clr_ale();
    }    
  }

	if (cmd == NAND_CMD_NONE)
		return;
    
  nand_write_byte(mtd, cmd);
}

/*
 *	read device ready pin
 *  @return 0 = busy, 1 = ready
 */
#if 0
static int nand512w3a_device_ready(struct mtd_info *mtd)
{
  return(0);
}
#endif

/*
 * Main initialization routine
 */
static int __init nand512w3a_init (void)
{
	struct nand_chip *this;
	const char *part_type = 0;
	void __iomem *nandaddr;

  //allocate ioport for control signals
  request_mem_region((unsigned long)na_nand_flash_pio,sizeof(np_pio),"pio_nand_flash");

	nandaddr = ioremap(na_nand_flash, 1024);
	if (!nandaddr) {
		printk("Failed to ioremap nand flash.\n");
		return -ENOMEM;
	}

	/* Allocate memory for MTD device structure and private data */
	nand512w3a_nand_mtd = kmalloc(sizeof(struct mtd_info) +
			     sizeof(struct nand_chip),
			     GFP_KERNEL);
	if (!nand512w3a_nand_mtd) {
		printk("Unable to allocate nand512w3a NAND MTD device structure.\n");
		iounmap ((void *) nandaddr);
		return -ENOMEM;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *) (&nand512w3a_nand_mtd[1]);

	/* Initialize structures */
	memset((char *) nand512w3a_nand_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	nand512w3a_nand_mtd->priv = this;
	nand512w3a_nand_mtd->owner = THIS_MODULE;

	/* insert callbacks */
	this->IO_ADDR_R = nandaddr;
	this->IO_ADDR_W = nandaddr;
	this->cmd_ctrl = nand512w3a_hwcontrol;
	this->dev_ready = NULL;	/* unknown whether that was correct or not so we will just do it like this */
	/* 15 us command delay time */
	this->chip_delay = 50;
	this->ecc.mode = NAND_ECC_SOFT;
	//this->eccmode = NAND_ECC_NONE;
	//this->options = NAND_NO_AUTOINCR;

	/* Scan to find existence of the device */
	if (nand_scan (nand512w3a_nand_mtd, 1)) {
		printk(KERN_NOTICE "No NAND device - returning -ENXIO\n");
		kfree (nand512w3a_nand_mtd);
		iounmap ((void *) nandaddr);
		return -ENXIO;
	}

	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(nand512w3a_nand_mtd, partition_info, NUM_PARTITIONS);

	/* Return happy */
	return 0;
}
module_init(nand512w3a_init);

/*
 * Clean up routine
 */
static void __exit nand512w3a_cleanup (void)
{
	struct nand_chip *this = (struct nand_chip *) &nand512w3a_nand_mtd[1];

	/* Release resources, unregister device */
	nand_release (nand512w3a_nand_mtd);

	/* Release io resource */
	iounmap ((void *) this->IO_ADDR_W);

	/* Free the MTD device structure */
	kfree (nand512w3a_nand_mtd);
  release_mem_region((unsigned long)na_nand_flash_pio,sizeof(np_pio));
}
module_exit(nand512w3a_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Manfred Schmid <manfred at wolfvision dot com>");
MODULE_DESCRIPTION("NAND flash driver for i12");
